# Copyright (C) 2022 Chi-kwan Chan
# Copyright (C) 2022 Steward Observatory
#
# This file is part of fadge.
#
# Fadge is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Fadge is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fadge.  If not, see <http://www.gnu.org/licenses/>.


from jax import numpy as np

from .bounds import *


def PHI(a, r):
    "Teo (2003) equation 11b"
    if a == 0 and r == 3:
        return 0 # == - (9/2) * (r-3)/a
    elif a == 1:
        return - (r*r - 2*r - 1)
    else:
        return - (r*r*r - 3*r*r + a*a*r + a*a) / (a * (r-1))

def Q(a, r):
    "Teo (2003) equation 11b"
    if a == 0 and r == 3:
        return 27 # == - ((9/2) * (r-3)/a)**2 + 27
    if a == 1:
        return - r*r*r * (r-4)
    else:
        return - r*r*r * (r*r*r - 6*r*r + 9*r - 4*a*a) / (a*a * (r-1)*(r-1))


def shadow(aspin=1.1, inc=np.pi/2):
    if abs(aspin) <= 1:
        r1 = rph1(aspin)
    else:
        r1 = (aspin*aspin - 1)**(1/3) + 1

    rs = np.linspace(r1+1e-6, rph2(aspin)-1e-6, num=100)
    if abs(aspin) > 1:
        rs = np.concatenate([np.array([np.nan]), rs])

    a = np.array([PHI(aspin, r) / np.sin(inc) for r in rs])
    b = np.array([np.sqrt(Q(aspin, r) + (aspin * np.cos(inc))**2 - (PHI(aspin, r) / np.tan(inc))**2) for r in rs])

    return (
        np.concatenate([-a, -a[::-1], -a[:1]]),
        np.concatenate([ b, -b[::-1],  b[:1]]),
    )



# Copyright (C) 2020 Chi-kwan Chan
# Copyright (C) 2020 Steward Observatory
#
# This file is part of fadge.
#
# Fadge is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Fadge is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fadge.  If not, see <http://www.gnu.org/licenses/>.


from jax import numpy as np, jit


def Cartesian(ndim=3, **kwargs):

    assert ndim > 0

    g = np.identity(ndim, **kwargs) # render constant metric

    def metric(x): # closure on `g`
        return g

    return metric


def Minkowski(ndim=4, **kwargs):

    assert ndim > 1

    g = np.diag(np.array([-1.0] + [1.0] * (ndim-1), **kwargs)) # render constant metric

    def metric(x): # closure on `g`
        return g

    return metric


def KerrSchild(aspin=0.0, ndim=4, **kwargs):

    assert ndim == 4

    eta = Minkowski(ndim)(None)
    aa  = aspin * aspin

    @jit
    def metric(x): # closure on `eta`, `aspin`, and `aa`
        zz = x[3] * x[3]
        kk = 0.5 * (x[1] * x[1] + x[2] * x[2] + zz - aa)
        rr = np.sqrt(kk * kk + aa * zz) + kk
        r  = np.sqrt(rr)
        f  = (2.0 * rr * r) / (rr * rr + aa * zz)
        l  = np.array([
            1.0,
            (r * x[1] + aspin * x[2]) / (rr + aa),
            (r * x[2] - aspin * x[1]) / (rr + aa),
            x[3] / r,
        ])
        return eta + f * l[:,np.newaxis] * l[np.newaxis,:]

    return metric



# Copyright (C) 2020 Chi-kwan Chan
# Copyright (C) 2020 Steward Observatory
#
# This file is part of PRay.
#
# PRay is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PRay is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PRay.  If not, see <http://www.gnu.org/licenses/>.


from xaj import odeint

from jax              import numpy as np
from jax.numpy        import dot
from jax.numpy.linalg import inv

from jax import jacfwd
from jax import jit


def JA(metric, ind=None):
    """Jacobian-Affine Formulation"""

    dmetric = jacfwd(metric) # "render" the Jacobian of the metric function

    @jit
    def rhs(state):
        x  = state[0]
        v  = state[1]

        g  =  metric(x)
        dg = dmetric(x)

        a  = inv(g) @ (- (dg @ v) @ v + 0.5 * v @ (v @ dg))
        if ind == 'time':
            return np.array([v, a - a[0] * v])
        else: # the default
            return np.array([v, a])

    return rhs


class Geode:

    def __init__(self, metric, l, s, ind=None, L=None, eqax=None, **kwargs):
        assert s.ndim >= 2

        rhs = lambda l, s: JA(metric, ind=ind)(s)
        if eqax is None:
            for offset in range(s.ndim-2+1):
                if s.shape[offset:offset+2] == (2,4):
                    eqax = [offset, offset+1]
                    break
            else:
                raise ValueError(f'cannot deduce `eqax` from state array with shape {s.shape}')

        kwargs['eqax'] = eqax
        if 'h' not in kwargs:
            kwargs['h'] = 1 if L is None else abs(L)
        self.geode = odeint(rhs, l, s, **kwargs)

        if L is not None:
            self.geode.extend(L)

    def extend(self, L, **kwargs):
        self.geode.extend(L, **kwargs)

    @property
    def lambdas(self):
        return self.geode.xs

    @property
    def states(self):
        return self.geode.ys

    def __call__(self, *args, **kwargs):
        return self.geode(*args, **kwargs)



# Copyright (C) 2022 Chi-kwan Chan
# Copyright (C) 2022 Steward Observatory
#
# This file is part of fadge.
#
# Fadge is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Fadge is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fadge.  If not, see <http://www.gnu.org/licenses/>.


from math  import sqrt
from cmath import cos, acos

from jax   import numpy as np


def _cos23acos(x):
    y = cos(2 * acos(x) / 3)
    if y.imag != 0:
        return np.nan
    else:
        return y.real

def cos23acos(X):
    try:
        return np.array([_cos23acos(x) for x in np.array(X)])
    except TypeError:
        return _cos23acos(X)


Z1 = lambda a: 1 + np.cbrt(1 - a*a) * (np.cbrt(1 + a) + np.cbrt(1 - a))
Z2 = lambda a: np.sqrt(3 * a * a + Z1(a) * Z1(a))

reh1 = lambda a: 1 - np.sqrt(1 - a * a) # inner event horizon
reh2 = lambda a: 1 + np.sqrt(1 - a * a) # outer event horizon
rph1 = lambda a: 2 * (1 + cos23acos(-abs(a))) # inner bound of photon orbits
rph2 = lambda a: 2 * (1 + cos23acos( abs(a))) # outer bound of photon orbits
rmb1 = lambda a: 2 - abs(a) + 2 * np.sqrt(1 - abs(a)) # inner marginally bounded orbits
rmb2 = lambda a: 2 + abs(a) + 2 * np.sqrt(1 + abs(a)) # outer marginally bounded orbits
rms1 = lambda a: 3 + Z2(a) - np.sqrt((3 - Z1(a)) * (3 + Z1(a) + 2 * Z2(a))) # inner stable circular orbits
rms2 = lambda a: 3 + Z2(a) + np.sqrt((3 - Z1(a)) * (3 + Z1(a) + 2 * Z2(a))) # outer stable circular orbits



# Copyright (C) 2020,2022 Chi-kwan Chan
# Copyright (C) 2020,2022 Steward Observatory
#
# This file is part of fadge.
#
# Fadge is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Fadge is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fadge.  If not, see <http://www.gnu.org/licenses/>.


from jax import numpy as np
from jax.experimental.maps import xmap

from .metric import KerrSchild
from .geode  import Geode
from .utils  import Nullify
from .icond  import cam, sphorbit


class GRRT:

    def __init__(self, aspin=0, hp=False, dtype=np.float32, **kwargs):
        self.aspin   = aspin
        self.dtype   = dtype
        self.kwargs  = kwargs

        self.metric  = KerrSchild(aspin)
        self.nullify = Nullify(self.metric)

        self.reh     = np.nan
        self._geode  = None
        self._ic     = None

        aa = self.aspin * self.aspin
        if aa <= 1:
            reh = 1.0 + np.sqrt(1 - aa)
            print('Radius of outer event horizon:', reh)
            if hp:
                print('Horizon penetrating')
            else:
                self.reh = reh
        else:
            print('There is no event horizon')

    def set_cam(self, r_obs=1e4, i_obs=60, j_obs=0):
        self.rij    = np.array([r_obs, np.radians(i_obs), np.radians(j_obs)], dtype=self.dtype)
        self.kwargs = {'L':-2*r_obs, 'h':0.75*r_obs, **self.kwargs}

    def set_pixels(self, a, b):
        def ic(ab): # closure on self.rij and self.nullify
            s = cam(self.rij, ab)
            return np.array([s[0], self.nullify(s[0],s[1])], dtype=ab.dtype)
        ab = np.array([a, b], dtype=self.dtype)
        self._ic = xmap(
            ic,
            in_axes ={i  :i for i in range(1,ab.ndim)},
            out_axes={i+1:i for i in range(1,ab.ndim)},
        )(ab)

    def set_image(self, fov=16, n=32, alpha0=0, beta0=0):
        axes = fov * ((np.arange(n) + 0.5) / n - 0.5)
        a, b = np.meshgrid(axes+alpha0, axes+beta0, indexing='ij')
        self.set_pixels(a, b)

    def set_axis(self, fov=16, n=32, PA=90, alpha0=0):
        r = fov * ((np.arange(n) + 0.5) / n - 0.5) + alpha0
        a = r * np.sin(PA * np.pi / 180)
        b = r * np.cos(PA * np.pi / 180)
        self.set_pixels(a, b)

    def set_ring(self, r=5.2, n=32):
        phi = 2 * np.pi * np.arange(n) / n
        a   = r * np.cos(phi)
        b   = r * np.sin(phi)
        self.set_pixels(a, b)

    def set_sphorbit(self, r=3):
        s = sphorbit(self.aspin, r)
        self._ic    = np.array([s[0], self.nullify(s[0],s[1])])
        self.kwargs = {'L':100, 'h':1, **self.kwargs}

    def geode(self, L=None, N=None, **kwargs):

        if self._ic is not None:
            kwargs = {'dtype':self.dtype, **self.kwargs, **kwargs} # compose kwargs
            if L is not None:
                kwargs.pop('L')

            aa = self.aspin * self.aspin
            def KSr(x): # closure on aa
                zz = x[3] * x[3]
                kk = 0.5 * (x[1] * x[1] + x[2] * x[2] + zz - aa)
                rr = np.sqrt(kk * kk + aa * zz) + kk
                return np.sqrt(rr)

            absaspin = abs(self.aspin)
            def KSd(x): # closure on absaspin
                dR = np.sqrt(x[1] * x[1] + x[2] * x[2]) - absaspin
                return np.sqrt(dR * dR + x[3] * x[3])

            fhupper = kwargs.pop('fhupper', 0.75)
            if 'hupper' not in kwargs:
                kwargs['hupper'] = lambda l, s: KSr(s[0]) * fhupper + 1

            fhlower = kwargs.pop('fhlower', None)
            if 'hlower' not in kwargs and fhlower is not None:
                kwargs['hlower'] = lambda l, s: fhlower

            eps = kwargs.pop('eps', 1e-2)
            if 'filter' not in kwargs:
                if np.isnan(self.reh):
                    kwargs['filter'] = lambda l, s: KSd(s[0]) >= eps
                else:
                    kwargs['filter'] = lambda l, s: KSr(s[0]) >= self.reh + eps

            self._geode = Geode(self.metric, 0, self._ic, **kwargs)
            self._ic    = None
        elif len(kwargs) > 0:
            print('Warning: ignore `kwargs`')

        if L is None:
            L = self.kwargs['L']
        try:
            len(L)
        except: # L is a scalar
            self._geode.extend(L, N=N)
            return self._geode.lambdas, self._geode.states
        else:   # L is not a scalar
            return self._geode(L, N=N)



# Copyright (C) 2022 Chi-kwan Chan
# Copyright (C) 2022 Steward Observatory
#
# This file is part of fadge.
#
# Fadge is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Fadge is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fadge.  If not, see <http://www.gnu.org/licenses/>.


from jax import numpy as np

from .shadow import PHI, Q


def cam(rij, ab):

    ci, si = np.cos(rij[1]), np.sin(rij[1])
    cj, sj = np.cos(rij[2]), np.sin(rij[2])

    R0 = rij[0] * si - ab[1] * ci # cylindrical radius
    z  = rij[0] * ci + ab[1] * si
    y  = R0     * sj + ab[0] * cj
    x  = R0     * cj - ab[0] * sj

    return np.array([
        [0, x, y, z],
        [1, si * cj, si * sj, ci],
    ], dtype=ab.dtype)


def sphorbit(aspin, r0):

    def thetadot(a, r):
        return np.sqrt(Q(a, r))

    def phidot(a, r):
        return (2*r*a + r*(r-2) * PHI(a, r)) / (r*r + a*a - 2*r)

    R = np.sqrt(r0*r0 + aspin*aspin)
    return np.array([
        [0, R, 0, 0],
        [1, 0, R * phidot(aspin, r0), -r0 * thetadot(aspin, r0)],
    ])



# Copyright (C) 2020 Chi-kwan Chan
# Copyright (C) 2020 Steward Observatory
#
# This file is part of fadge.
#
# Fadge is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Fadge is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fadge.  If not, see <http://www.gnu.org/licenses/>.


from jax       import numpy as np
from jax.numpy import dot
from math      import isclose


def quadratic(A, b, C):
    bb = b * b
    AC = A * C
    dd = np.select([~np.isclose(bb, AC)], [bb - AC], 0)
    bs = np.heaviside(b, 1)
    D  = - (b + bs * np.sqrt(dd))
    x1 = D / A
    x2 = C / D
    return np.minimum(x1, x2), np.maximum(x1, x2)


def Nullify(metric, p=1):

    assert p > 0

    def nullify(x, v): # closure on `p`
        g = metric(x)
        A = v[:p] @ g[:p,:p] @ v[:p]
        b = v[p:] @ g[p:,:p] @ v[:p]
        C = v[p:] @ g[p:,p:] @ v[p:]

        d1, d2 = quadratic(A, b, C)
        S      = np.select([d1 > 0, d2 > 0], [d1, d2], np.nan)

        return np.concatenate([v[:p], v[p:] / S])

    return nullify



#!/usr/bin/env python3
#
# Copyright (C) 2020 Chi-kwan Chan
# Copyright (C) 2020 Steward Observatory
#
# This file is part of fadge.
#
# Fadge is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Fadge is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fadge.  If not, see <http://www.gnu.org/licenses/>.


from setuptools import setup, find_packages


setup(
    name='fadge',
    version='0.1.5',
    url='https://github.com/adxsrc/fadge',
    author='Chi-kwan Chan',
    author_email='chanc@arizona.edu',
    description='Fast Automatic Differential GEometry',
    packages=find_packages('mod'),
    package_dir={'': 'mod'},
    entry_points={
        'console_scripts': [
            'fadge = fadge.__main__:fadge',
        ],
    },
    python_requires='>=3.7',
    install_requires=[
        'click>=7.1.2',
        'h5py',
        'xaj>=0.1.6,<0.2',
    ],
)



# Copyright (C) 2022 Chi-kwan Chan
# Copyright (C) 2022 Steward Observatory
#
# This file is part of fadge.
#
# Fadge is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Fadge is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fadge.  If not, see <http://www.gnu.org/licenses/>.


from jax import numpy as np
try:
    from collections import Iterable
except:
    Iterable = (tuple, list)


def KSHorizons(aspin, n=None):

    if n is None:
        n = 46 if aspin <= 1 else 721

    phi   = np.linspace(-np.pi, np.pi, 2*(n-1)+1)
    theta = np.linspace(0, np.pi, n)
    ct    = np.cos(theta)
    st    = np.sin(theta)

    aa    = aspin * aspin
    rhp   = 1 + np.sqrt(1 - aa)
    rhm   = 1 - np.sqrt(1 - aa)
    rep   = 1 + np.sqrt(1 - aa * ct * ct)
    rem   = 1 - np.sqrt(1 - aa * ct * ct)

    def Rz(r): # work for both scalar and array `r`
        R  = st * np.sqrt(r * r + aa)
        z  = ct * r
        return R[np.isfinite(R)], z[np.isfinite(z)]

    def plot2(ax, R, z, *args, **kwargs):
        ax.plot( R, z, *args, **kwargs)
        ax.plot(-R, z, *args, **kwargs)

    def xy(r):
        R = max(np.sqrt(r * r + aa), 1e-4)
        return R * np.sin(phi), R * np.cos(phi)

    def plotedgeon(ax, color=None, eh=True, es=True, r=[]):
        p = ax.plot([-aspin, aspin], [0,0], ':', color=color)
        c = p[0].get_color()
        ax.scatter([-aspin, aspin], [0,0], color=c, label=f'a={aspin}')

        Rep, zep = Rz(rep)
        Rem, zem = Rz(rem)
        if abs(aspin) > 1:
            Re = np.concatenate([Rep, Rem[::-1], Rep[:1]])
            ze = np.concatenate([zep, zem[::-1], zep[:1]])
            if es:
                plot2(ax, Re, ze, color=c, linewidth=1)
        else:
            if es:
                plot2(ax, Rep, zep, color=c, linewidth=1)
                plot2(ax, Rem, zem, color=c, linewidth=1)
            if eh:
                plot2(ax, *Rz(rhp), color=c)
                plot2(ax, *Rz(rhm), color=c)

        for r0 in r:
            plot2(ax, *Rz(r0), '--', color=c, linewidth=1)

    def plotfaceon(ax, color=None, eh=True, es=True, r=[]):
        p = ax.plot(*xy(0.0), color=color, linewidth=4, label=f'a={aspin}')
        c = p[0].get_color()

        if abs(aspin) > 1:
            if es:
                ax.plot(*xy(2), color=c, linewidth=1)
        else:
            if es:
                ax.plot(*xy(2), color=c, linewidth=1)
            if eh:
                ax.plot(*xy(rhp), color=c)
                ax.plot(*xy(rhm), color=c)

        for r0 in r:
            ax.plot(*xy(r0), '--', color=c, linewidth=1)

    def horizons(ax=None, color=None, faceon=False, eh=True, es=True, r=None, **kwargs):
        if ax is None:
            from matplotlib import pyplot as plt
            fig, ax = plt.subplots(1,1,**kwargs)
            ax.set_aspect('equal')

        if r is None:
            r = []
        elif not isinstance(r, Iterable):
            r = [r]

        if faceon:
            plotfaceon(ax, color, eh, es, r)
        else:
            plotedgeon(ax, color, eh, es, r)

    return horizons



# Copyright (C) 2020 Chi-kwan Chan
# Copyright (C) 2020 Steward Observatory
#
# This file is part of fadge.
#
# Fadge is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Fadge is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fadge.  If not, see <http://www.gnu.org/licenses/>.


from . import bounds, geode, icond, metric, shadow, utils, viz
from .core import GRRT



# Copyright (C) 2022 Chi-kwan Chan
# Copyright (C) 2022 Steward Observatory
#
# This file is part of fadge.
#
# Fadge is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Fadge is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fadge.  If not, see <http://www.gnu.org/licenses/>.


from jax.config import config
config.update("jax_enable_x64", True)

import numpy as np
import h5py
import click

from fadge import GRRT


#==============================================================================
@click.group()
def fadge():
    pass


#==============================================================================
@fadge.command()

@click.option('--eps',         default=1e-3,    help="Stop integration `eps` outside the event horizon")
@click.option('--atol',        default=1e-6,    help="Absolute error tolerance in numerical integration")
@click.option('--setup',       default='image', help='Initial condition setup; can be "image" or "axis"')
@click.option('--alpha0',      default=0.0,     help='Initial condition offset along alpha direction')
@click.option('--beta0',       default=0.0,     help='Initial condition offset along beta  direction')
@click.option('--full/--ends', default=False,   help="Output full geodesics or only end points")

@click.argument('aspin',       type=float)
@click.argument('inclination', type=float)

def grrt(aspin, eps, setup, alpha0, beta0, full, atol, inclination):
    print( "Fadge: general relativistic ray tracing")
    print(f"    aspin       = {aspin:.2f}")
    print(f"    inclination = {inclination:g}")

    ns = GRRT(
        aspin, ind='time',
        eps=eps, atol=atol, rtol=0, hmin=0,
        names={'ind':'t'},
        dtype=np.float64,
        steps=full, dense=False,
    )

    ns.set_cam(1e4, inclination, 0)
    if setup == 'image':
        ns.set_image(16, 128, alpha0=alpha0, beta0=beta0)
        out = f'image_a{aspin:.2f}_i{inclination:g}.h5'
    elif setup == 'axis':
        if beta0 != 0.0:
            print('WARNING: set_axis() does not support setting `beta0`; ignore')
        ns.set_axis(16, 1024, alpha0=alpha0)
        out = f'axis_a{aspin:.2f}_i{inclination:g}.h5'
    else:
        raise ValueError(f'Unknown setup "{setup}".')

    l, f = ns.geode()

    with h5py.File(out, 'w') as h:
        h.create_dataset('l', data=np.array(l))
        h.create_dataset('f', data=np.array(f))


#==============================================================================
# In case run as a script
if __name__ == "__main__":
    fadge()
